/*****************************************************************************
* IrtDspList.h - header file for the IrtDspListClass class.		     *
* This class implements a list.						     *
******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                *
******************************************************************************
* Written by Eran Karpen and Gershon Elber, July 2006.			     *
*****************************************************************************/

#ifndef IRT_DSP_LIST_H
#define IRT_DSP_LIST_H

#include <stdio.h>
#include <assert.h>


#define IRT_DSP_LIST_DEF_SIZE	5

#define IRT_DSP_FOR_EACH(List_, ItemType_) \
    for (unsigned int _i = 0; _i < (List_).Size(); _i++) { \
	ItemType_ &Item = (List_)[_i];


#define IRT_DSP_FOR_EACH_ITEM(List_, ItemType_, Func_) { \
    for (unsigned int _i = 0; _i < (List_).Size(); _i++) { \
	ItemType_ &Item = (List_)[_i]; \
        Func_; \
    } \
};

#define IRT_DSP_FOR_EACH_ITEM_RVRSD(List_, ItemType_, Func_) { \
    for (unsigned int _i = (List_).Size() - 1; _i >= 0; _i--) { \
	ItemType_ &Item = (List_)[_i]; \
	Func_; \
    } \
};

template<class T>
class IrtDspListClass {
    public:
	IrtDspListClass<T>(unsigned int StartSize = IRT_DSP_LIST_DEF_SIZE) {
	    Vec = new T[AllocSize = StartSize];
	    NumItems = 0;
	};
	IrtDspListClass<T>(const IrtDspListClass<T> &Val) {
	    Vec = new T[AllocSize = Val.AllocSize];
	    IRIT_GEN_COPY(Vec, Val.Vec, sizeof(T) * (NumItems = Val.NumItems));
	}
	~IrtDspListClass<T>() {
	    delete [] Vec;
	}
	IrtDspListClass<T> &operator=(const IrtDspListClass<T> &Val) {
	    delete [] Vec;
	    Vec = new T[AllocSize = Val.AllocSize];
	    IRIT_GEN_COPY(Vec, Val.Vec,  sizeof(T) * (NumItems = Val.NumItems));
	    return *this;
	}
        void Clear(bool DeleteItems = false) {
	    if (DeleteItems) {
		unsigned int i;
		for (i = 0; i < NumItems; i++)
		    delete Vec[i];
	    }
	    NumItems = 0;
	}
        bool Empty() const { return NumItems == 0; }
        unsigned int Size() const { return NumItems; }
	T &operator[](unsigned int Idx) {
	    assert(Idx < NumItems);
	    return Vec[Idx];
	}
	const T &operator[](unsigned int Idx) const {
	    assert(Idx < NumItems);
	    return Vec[Idx];
	}
        bool Delete(const T &Val) {
	    unsigned int i;
	    for (i = 0; i < NumItems; i++) {
	        if (Val == Vec[i]) {
		    NumItems--;
		    for ( ; i < NumItems; i++)
		        Vec[i] = Vec[i + 1];
		    return true;
		}
	    }
	    return false;
	}
	T Remove(unsigned int Idx) {
	    assert(Idx < NumItems);
            T result = Vec[Idx];
            if (--NumItems > 0) {
	        unsigned int i;
	        for (i = Idx; i < NumItems; i++)
	            Vec[i] = Vec[i + 1];
	    }
            return result;
	}
	bool Exist(const T &Val) const {
	    unsigned int i;
	    for (i = 0; i < NumItems; i++) {
	        if (Val == Vec[i])
		    return true;
	    }
	    return false;
	}
	T &Get(unsigned int Idx) {
	    assert(Idx < NumItems);
	    return Vec[Idx];
	}
	T &GetLast() { return Get(Size() - 1); }
	T &GetFirst() { return Get(0); }
	T *Pop() {
	    if (Empty())
		return NULL;
	    return &Vec[--NumItems];	    
	}
	void Append(const T &Item) {
	    if (NumItems >= AllocSize) {
	        T *NewVec = new T[AllocSize *= 2];
		memcpy(NewVec, Vec,
		       sizeof(T) * (NumItems < AllocSize ? NumItems : AllocSize));
		delete [] Vec;
		Vec = NewVec;
	    }
	    Vec[NumItems++] = Item;
	}
	void Insert(const T &Item, unsigned int Idx) {
	    assert(Idx <= NumItems);
	    if (Idx == NumItems)
		Append(Item);
	    else {
		if (NumItems + 1 >= AllocSize - 1) {
		    T *NewVec = new T[AllocSize * 2];
		    memcpy(NewVec, Vec,  sizeof(T) * NumItems);
		    delete [] Vec;
		    Vec = NewVec;
		    AllocSize *= 2;
		}
		for (unsigned int i = NumItems; i > Idx; i--)
		    Vec[i] = Vec[i - 1];
		NumItems++;
		Vec[Idx] = Item;
	    }
	}
 	void Add(IrtDspListClass &AList) {
	    IRT_DSP_FOR_EACH_ITEM(AList, T, Append(Item));
 	}
	bool CompareItems(IrtDspListClass &AList) const {
	    if (NumItems != AList.Size())
		return false;
	    IRT_DSP_FOR_EACH_ITEM(AList, T,
			          if (!Exist(Item))
			              return false; );
	    return true;
	}
	void SwapItems(const T &First, const T &Second) {
	    unsigned int i, 
		FirstIdx = NumItems, 
		SecondIdx = NumItems;

	    for (i = 0; i < NumItems; ++i) {
		if (Vec[i] == First)
		    FirstIdx = i;
		else if (Vec[i] == Second)
		    SecondIdx = i;
		if (FirstIdx < NumItems && SecondIdx < NumItems)
		    break;
	    }
	    SwapItems(FirstIdx, SecondIdx);
	}
	void SwapItems(unsigned int FirstIdx, unsigned int SecondIdx) {
	    T Temp;

	    if (FirstIdx >= NumItems || SecondIdx >= NumItems)
		return;
	    if (FirstIdx == SecondIdx)
		return;
	    Temp = Vec[FirstIdx];
	    Vec[FirstIdx] = Vec[SecondIdx];
	    Vec[SecondIdx] = Temp;
	}

    private:
	T *Vec;
	unsigned int NumItems, AllocSize;
};

#endif /* IRT_DSP_LIST_H */
