/*****************************************************************************
* This header file contains all type declarations required in order to use   *
* this library. In particular, all functions (and only those) provided for   *
* external usage are declared in this header file. Such functions available  *
* for external use start with the prefix "GC_", while functions without this *
* prefix are library-internal functions and are not to be called from        *
* outside the library.                                                       *
*                                                                            *
* IMPORTANT: If a graphics card driver error occurs, try decreasing the      *
* value of the variable NrOfIterations.                                      *
*****************************************************************************/

#ifndef GC_H
#define GC_H

/****************************************************************************/

#include <CL/opencl.h>

/*****************************************************************************
*** Definitions **************************************************************
*****************************************************************************/

#define GC_FALSE                               0
#define GC_TRUE                                1

#define GC_SPEED_UP_NO_APPROXIMATION           (-1)
#define GC_OCCLUDER_TRIANGLE_ID                0
#define GC_MIN_TARGET_TRIANGLE_ID              (GC_OCCLUDER_TRIANGLE_ID + 1)

#define GC_MATRIX_DIMENSION                    4

#define GC_MAX_NEIGHBORHOOD_RADIUS             6

/*****************************************************************************
*** Enumerations *************************************************************
******************************************************************************
*/                                                                          /*
*/  typedef enum {                                                          /*
*/                                                                          /*
*/      GC_BIT_PRECISION,                                                   /*
*/      GC_BYTE_PRECISION,                                                  /*
*/      GC_SHORT_PRECISION,                                                 /*
*/      GC_INT_PRECISION                                                    /*
*/                                                                          /*
*/  } GC_PrecisionType;                                                     /*
*/                                                                          /*
*/  typedef enum { GC_EXHAUSTIVE, GC_GREEDY } GC_AlgorithmType;             /*
*/                                                                          /*
*/  typedef enum { GC_INFO, GC_WARNING, GC_ERROR } GC_MessageKind;          /*
*/                                                                          /*
*/  typedef enum { GC_XYZ, GC_UV, GC_NR_OF_SPACES } GC_Space;               /*
*/                                                                          /*
*/  typedef enum {                                                          /*
*/                                                                          /*
*/      GC_STORED_UV_MAP_FILE,                                              /*
*/      GC_TARGET_UV_MAP_FILE,                                              /*
*/      GC_BEST_N_UV_MAPS_OF_FILE     /* Has to remain last entry.           *
*/                                                                          /*
*/  } GC_FileTypeEnum;                                                      /*
*/                                                                          /*
*/  typedef enum {                                                          /*
*/                                                                          /*
*/      GC_VISUALIZATION_TYPE_RAW,                                          /*
*/      GC_VISUALIZATION_TYPE_NORMALIZED_IF_REQUIRED,                       /*
*/      GC_VISUALIZATION_TYPE_NORMALIZED_ACCORDING_TO_MAX_VALUE,            /*
*/      GC_VISUALIZATION_TYPE_NORMALIZED_ACCORDING_TO_MAX_TARGET_VALUE,     /*
*/      GC_VISUALIZATION_TYPE_5_COLOR_MAP,                                  /*
*/      GC_VISUALIZATION_TYPE_10_COLOR_MAP,                                 /*
*/      GC_VISUALIZATION_TYPE_DEVIATION_MULTIPLICATIVE,                     /*
*/      GC_VISUALIZATION_TYPE_DEVIATION_ADDITIVE,                           /*
*/                                                                          /*
*/  } GC_VisualizationTypeEnum;                                             /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Platform dependent type definitions **************************************
*****************************************************************************/

#ifdef WIN32
    typedef cl_int  GC_LargestSignedIntegerType;
    typedef cl_uint GC_LargestUnsignedIntegerType;
#else /* 64 bit system */
    typedef cl_long  GC_LargestSignedIntegerType;
    typedef cl_ulong GC_LargestUnsignedIntegerType;
#endif

/*****************************************************************************
*** Platform independent type definitions ************************************
******************************************************************************
*/                                                                          /*
*/  typedef cl_double               GC_1DVectorType;                        /*
*/  typedef cl_double               GC_2DVectorType[2];                     /*
*/  typedef cl_double               GC_3DVectorType[3];                     /*
*/  typedef cl_double               GC_4DVectorType[4];                     /*
*/                                                                          /*
*/  typedef cl_uchar                GC_BooleanType;                         /*
*/  typedef cl_uint                 GC_ImageSizeType;                       /* Type indicating largest value allowed for image dimension.
*/  typedef cl_uchar                GC_BitIteratorType;                     /* Type large enough to count the size of types in bits.
*/  typedef cl_uchar                GC_DeviceVectorWorkItemCardinalityType; /* Type used for storing the vector cardinality of the OpenCL workitems.
*/                                                                          /*
*/  typedef cl_char                 GC_CharacterType;                       /* Type used for characters. 
*/                                                                          /*
*/  typedef cl_uint                 GC_UVMapChunkType;                      /* Type used for chunk of UV map values.
*/                                                                          /*
*** Types used only in GC_Renderer *******************************************
*/                                                                          /*
*/  typedef cl_int                          GC_NrOfSensorsType;             /* Type used for storing number of sensors.
*/  typedef GC_LargestUnsignedIntegerType   GC_NrOfObjectsType;             /* Type used for storing number of objects (occluder and target objects).
*/                                                                          /*
*/  typedef cl_uint                         GC_XYZMapValueType;             /* Type used for triangle IDs (= values used in XYZ maps).
*/  typedef GC_LargestUnsignedIntegerType   GC_NrOfVerticesType;            /* Type used for number of vertices.
*/  typedef GC_3DVectorType                 GC_SensorPositionType;          /* Type used for position of sensor.
*/  typedef GC_3DVectorType                 GC_VerticesXYZType;             /* Type used for coordinates of vertices in Euclidean 3D space.
*/  typedef GC_2DVectorType                 GC_VerticesUVType;              /* Type used for coordinates of vertices in parametric space.
*/  typedef GC_3DVectorType                 GC_NormalsType;                 /* Type used for normals.
*/  typedef cl_float                        GC_UVMapScalingFactor;          /* Type used for scaling factor of UV map values (see also variable UVMapScalingFactor in GC_RendererParametersStruct).
*/  typedef cl_uint                         GC_NeighborhoodRadiusType;      /* Type used for neighborhood radius (see also variable NeighborhoodRadius in GC_RendererParametersStruct).
*/                                                                          /*
*** Types used only in GC_CoverageSolver *************************************
*/                                                                          /*
*/  typedef cl_int                          GC_InputVectorIndexType;        /*
*/  typedef GC_LargestSignedIntegerType     GC_NrOfWeightsType;             /*
*/                                                                          /*
*/  typedef cl_uint                         GC_NrOfIterationsType;          /*
*/  typedef cl_double                       GC_CoverageValueType;           /*
*/  typedef cl_uint                         GC_WorkGroupSizeType;           /*
*/  typedef cl_uint                         GC_BandWidthType;               /*
*/  typedef cl_double                       GC_FactorType;                  /*
*/  typedef cl_uint                         GC_NrOfFoundSolutionsType;      /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Union for number of sensors / number of input vectors ********************
******************************************************************************
*/                                                                          /*
*/  typedef union {                                                         /*
*/                                                                          /*
*/      GC_InputVectorIndexType Sensors;                                    /*
*/      GC_InputVectorIndexType InputVectors;                               /*
*/                                                                          /*
*/  } GC_NrOfUnion;                                                         /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** UV map structure *********************************************************
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_UVMapChunkType*            Data;                                 /* Serialized and concatenated UV maps of all objects
*/      GC_LargestUnsignedIntegerType NrOfUVMapValues;                      /*
*/                                                                          /*
*/  } GC_UVMapsStruct;                                                      /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Matrix structure *********************************************************
******************************************************************************
*/                                                                          /*
*/  typedef GC_LargestUnsignedIntegerType   GC_MatrixDimensionType;         /*
*/  typedef cl_double                       GC_MatrixEntryType;             /*
*/                                                                          /*
*** Struct *******************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_MatrixDimensionType RowDimension;                                /*
*/      GC_MatrixDimensionType ColumnDimension;                             /*
*/      GC_MatrixEntryType*    Entries;                                     /*
*/                                                                          /*
*/  } GC_MatrixStruct;                                                      /*
*/                                                                          /*
*** Functions ****************************************************************
*/                                                                          /*
*/  GC_MatrixStruct* GC_CopyMatrix                                          /*
*/      (const GC_MatrixStruct* MatrixToCopy);                              /*
*/                                                                          /*
******************************************************************************
*/                                                                          /*
*   Description:                                                             *
*     GC_CreateMatrix returns a matrix structure with RowDimension rows and  *
*     ColumnDimensions columns. The memory required for storing the matrix   *
*     values is allocated by the function itself and has to be freed by      *
*     using the function GC_DeleteMatrix. If input parameter Entries is      *
*     NULL, all matrix entries are set to 0. Otherwise, the values of        *
*     Entries are copied. In this case, Entries has to be a one-dimensional  *
*     array of the form A_{0,0}, A_{0,1}, ..., A_{0,ColumnDimension-1},      *
*     A_{1,0}, A_{1,1}, ..., A_{1,ColumnDimension-1}, ...,                   *
*     A_{RowDimension-1,0}, A_{RowDimension-1,1}, ...,                       *
*     A_{RowDimension-1,ColumnDimension-1}, where A_{i,j} is the entry in    *
*     the i.th row and the j.th column of some matrix A.                     *
*/                                                                          /*
*/  GC_MatrixStruct* GC_CreateMatrix                                        /*
*/      (const GC_MatrixDimensionType RowDimension,                         /*
*/       const GC_MatrixDimensionType ColumnDimension,                      /*
*/       const GC_MatrixEntryType*    Entries);                             /*
*/                                                                          /*
******************************************************************************
*/                                                                          /*
*/  void GC_DeleteMatrix                                                    /*
*/      (      GC_MatrixStruct*       Matrix);                              /*
*/                                                                          /*
*/  void GC_ReplaceMatrixByIdentityMatrix                                   /*
*/      (      GC_MatrixStruct*       Matrix);                              /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Sensor structure *********************************************************
******************************************************************************
*   Type large enough to cover the integer range {0, 1, ...,                 *
*   GC_MAX_NR_OF_VIEWS}.                                                     *
*/  typedef cl_uchar GC_NrOfViewsType;                                      /*
*/                                                                          /*
*** Struct *******************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_SensorPositionType Position;                                     /*
*/      GC_NrOfViewsType      NrOfViews;                                    /*
*/      GC_MatrixStruct**     ProjectionMatrices;                           /*
*/      GC_MatrixStruct**     InverseTransformationMatrices;                /*
*/                                                                          /*
*/      GC_CharacterType*     Name;                                         /* Dynamically allocated array. Will be released by function DestroyGeneralParameters
*/                                                                          /*
*/  } GC_SensorStruct;                                                      /*
*/                                                                          /*
*** Functions ****************************************************************
*/                                                                          /*
*/  GC_SensorStruct* GC_CreateSensor                                        /*
*/      (const GC_SensorPositionType    Position,                           /*
*/       const GC_NrOfViewsType         NrOfViews,                          /*
*/       const GC_MatrixStruct**        ProjectionMatrices,                 /*
*/       const GC_MatrixStruct**        InverseTransformationMatrices,      /*
*/       const GC_CharacterType*        Name);                              /*
*/                                                                          /*
******************************************************************************
*/                                                                          /*
*/  void GC_DestroySensor                                                   /*
*/      (      GC_SensorStruct*         Sensor);                            /*
*/                                                                          /*
******************************************************************************



/*****************************************************************************
*** Object structure *********************************************************
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_NrOfVerticesType  NrOfVertices;                                  /*
*/      GC_VerticesXYZType*  VerticesXYZ;                                   /* XYZ world coordinates of the object
*/      GC_VerticesUVType*   VerticesUV;                                    /* UV coordinates of the object
*/      GC_NormalsType*      Normals;                                       /*
*/      GC_XYZMapValueType*  IDs;                                           /*
*/                                                                          /*
*/      GC_MatrixStruct*     TransformationMatrix;                          /*
*/                                                                          /*
*/      GC_ImageSizeType     UVImageWidth;                                  /*
*/      GC_ImageSizeType     UVImageHeight;                                 /*
*/                                                                          /*
*/      GC_CharacterType*    Name;                                          /* Dynamically allocated array. Will be released by function DestroyGeneralParameters.
*/                                                                          /*
*/  } GC_ObjectStruct;                                                      /*
*/                                                                          /*
*** Functions ****************************************************************
*/                                                                          /*
*/  GC_ObjectStruct* GC_CreateObject                                        /*
*/      (const GC_NrOfVerticesType  NrOfVertices,                           /*
*/       const GC_VerticesXYZType*  VerticesXYZ,                            /*
*/       const GC_VerticesUVType*   VerticesUV,                             /*
*/       const GC_NormalsType*      Normals,                                /*
*/       const GC_XYZMapValueType*  IDs,                                    /*
*/       const GC_MatrixEntryType*  TransformationMatrixEntries,            /*
*/       const GC_CharacterType*    Name);                                  /*
*/                                                                          /*
*/  void GC_DeleteObject                                                    /*
*/      (      GC_ObjectStruct* Object);                                    /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Structure for general input data parameters ******************************
******************************************************************************
*** This structure contains the variables required for both the rendering  ***
*** of the XYZ/UV maps and the coverage solving. The structure is linked   ***
*** to both the structure for input data parameters of the renderer and    ***
*** the structure for input data parameters of the coverage solver.        ***
*** After execution of the code, the dynamically allocated memory of those ***
*** structures is freed, while this structure is re-linked to the          ***
*** structure containing the coverage solution, since it's values are      ***
*** required for visualization.                                            ***
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_BooleanType                LogToExtern;                          /* Indicates whether the log messages shall be passed to the print function provided by the external code using this library.
*/      GC_BooleanType                TalkativeLogging;                     /* Indicates whether more logging should be performed.
*/      GC_BooleanType                LogToFile;                            /* Indicates whether the log messages shall be stored in a file.
*/      GC_BooleanType                StrictConsistency;                    /* Indicates whether an error message shall be produced and the run aborted in case that a pixel has a Don't Care value in the rendered UV maps, but a required coverage value in the corresponding target map.
*/      GC_BooleanType                IsOptimizationRun;                    /* Indicates whether the current run is a test run for computing the optimal setttings.
*/      GC_BooleanType                OverwriteWorkDirectory;               /* Indicates whether the work directory shall be overwritten if it exists already.
*/      GC_BooleanType                DeleteStoredUVMapsFiles;              /* Indicates whether the stored UV maps files and the corresponding directory should be deleted.
*/                                                                          /*
*/      GC_PrecisionType              Precision;                            /* The used precision. Must be GC_BIT_PRECISION or GC_BYTE_PRECISION.
*/      GC_BooleanType                BinarySensors;                        /* Indicates whether the coverage values provided by the sensors are binary (for visibility and k-coverage) or not (for illumination).
*/                                                                          /*
*/      GC_NrOfObjectsType            NrOfTargets;                          /* The number of targets. Must be greater than 0 (if UV maps are rendered).
*/      GC_LargestUnsignedIntegerType NrOfOccluders;                        /* The number of occluders.
*/      GC_NrOfUnion                  NrOf;                                 /* The number of sensors/input vectors. Must be greater than 0.
*/                                                                          /*
*/      GC_ImageSizeType              ImageWidth[GC_NR_OF_SPACES];          /* For XYZ maps: width of computed XYZ maps. For UV maps: maximal(!) width (among all objects) of computed UV maps. Must be multiple of 2.
*/      GC_ImageSizeType              ImageHeight[GC_NR_OF_SPACES];         /* For XYZ maps: height of computed XYZ maps. For UV maps: maximal(!) height (among all objects) of computed UV maps. Must be multiple of 2.
*/                                                                          /*
*/      GC_BooleanType                ReadTargetVector;                     /* Indicates whether the files containing the elements of the coverage requirement vector shall be read from the disk.
*/                                                                          /*
*/      GC_ObjectStruct**             Objects;                              /* The list of objects (both targets and occluders).
*/      GC_SensorStruct**             Sensors;                              /* The list of sensors.
*/                                                                          /*
*/      GC_CharacterType*             MainDirectoryPath;                    /* The (existing) main directory in which the working folder will be created. Dynamically allocated memory. Will be automatically released if no solution was found.
*/      GC_CharacterType*             WorkDirectoryPath;                    /* The working directory in which all data is stored. Must be a path to a (not necessarily existing) folder in MainDirectoryPath. Dynamically allocated memory. Will be automatically released if no solution was found.
*/      GC_CharacterType*             InputDirectoryPath;                   /* The directory in which the input target maps reside. Dynamically allocated memory. Will be automatically released if no solution was found.
*/                                                                          /*
*/      cl_double                     CoverageValueUpperBoundFactor;        /* A target location t is considered "correctly covered" iff l * (coverage value required at t) <= (current coverage value at t)\n <= u * (coverage value required at t), where l is the lower bound factor and u is the upper bound factor.
*/      cl_double                     CoverageValueLowerBoundFactor;        /* Please refer to the description of CoverageValueUpperBoundFactor.
*/                                                                          /*
*/  } GC_GeneralParametersStruct;                                           /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Structure for input data parameters of renderer **************************
******************************************************************************
*** This structure contains the variables solely required during the       ***
*** rendering of the XYZ/UV maps.                                          ***
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_GeneralParametersStruct* GeneralParameters;                      /*
*/                                                                          /*
*/      GC_BooleanType              ComputeImages;                          /* Indicates whether UV maps shall be created.
*/      GC_BooleanType              StoreAllImages[GC_NR_OF_SPACES];        /* Indicates whether the created images shall be stored to png files.
*/                                                                          /*
*/      GC_NeighborhoodRadiusType   NeighborhoodRadius;                     /* Radius of neighborhood of a pixel in the XYZ map to be checked in order to decide visibility of a pixel in the UV map.
*/                                                                          /*
*/      GC_UVMapScalingFactor       UVMapScalingFactor;                     /* The factor by which coverage levels of a sensor have to be scaled.
*/                                                                          /*
*/  } GC_RendererParametersStruct;                                          /*
*/                                                                          /*
******************************************************************************



/*****************************************************************************
*** Structure for performance optimization parameters of coverage solver *****
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_BooleanType                      UseCompressedInputVectors;      /* Indicates whether the vector collapsing should be performed.
*/      GC_BooleanType                      SortingSpeedUp;                 /* Indicates whether the weights used in the case of vector collapsing should be ordered, possibly leading to a performance improvement.
*/      GC_NrOfIterationsType               NrOfIterations;                 /* The number of sensor combinations for which the coverage value should be computed (on the graphics card).
*/                                                                          /*
*/      GC_BitIteratorType                  NrOfBitsInDeviceScalarWorkItem; /* The number of bits of the scalar work item used for the coverage value computation on the graphics card. Must be in {GC_NR_OF_BITS_IN_CL_UCHAR, GC_NR_OF_BITS_IN_CL_USHORT, GC_NR_OF_BITS_IN_CL_UINT} if used precision is GC_BIT_PRECISION and GC_NR_OF_BITS_IN_CL_UCHAR if used precision is GC_BYTE_PRECISION.
*/      GC_DeviceVectorWorkItemCardinalityType DeviceVectorWorkItemCardinality;/* The vector cardinality of the work item used for the coverage value computation on the graphics card. Must be in {1, 2, 4, 8, ..., GC_MAXIMUM_INPUT_ELEMENT_CARDINALITY}.
*/                                                                          /*
*/      GC_WorkGroupSizeType                WorkGroupSizeSumsKernel;        /* ^
*/      GC_WorkGroupSizeType                WorkGroupSizeResetAux0Kernel;   /* | The number of work items per work group for the employed kernels.
*/      GC_WorkGroupSizeType                WorkGroupSizeTargetSumKernel;   /* |
*/      GC_WorkGroupSizeType                WorkGroupSizeSubtract;          /* v
*/                                                                          /*
*/  } GC_CoverageOptimizationParametersStruct;                              /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Structure for input data parameters of coverage solver *******************
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_GeneralParametersStruct*   GeneralParameters;                    /* Has to be dynamically allocated.
*/                                                                          /*
*/      GC_BooleanType                ComputeCoverage;                      /* Indicates whether the coverage values shall be computed.
*/                                                                          /*
*/      GC_InputVectorIndexType       MinCardinality;                       /* The minimal cardinality of the solution set in case that the exhaustive algorithm is used.
*/      GC_InputVectorIndexType       MaxCardinality;                       /* The maximal cardinality of the solution set in case that the exhaustive algorithm is used.
*/                                                                          /*
*/      GC_NrOfFoundSolutionsType     MaxNrOfStoredSolutions;               /* The maximal number of stored solutions.
*/                                                                          /*
*/      GC_AlgorithmType              AlgorithmType;                        /* The type of the used algorithm. Must be GC_EXHAUSTIVE or GC_GREEDY.
*/                                                                          /*
*/      cl_double                     ApproximationFactor;                  /* A sensor combination is considered a valid solution if (nr of target locations correctly covered by sensor combination) >= a * (nr of coverable target locations), where a is the approximation factor specified in this field. For the meaning of "correctly covered" please refer to the description of CoverageValueUpperBoundFactor.
*/                                                                          /*
*/      GC_NrOfWeightsType            SpeedUpNrOfUsedExactValues;           /* The number of weights whose value is used without being rounded to the closest multiple of 2. Must be GC_SPEED_UP_NO_APPROXIMATION if no value shall be approximated and non-negative value else.
*/                                                                          /*
*/      GC_CoverageOptimizationParametersStruct Optimization;               /* See declaration of GC_CoverageOptimizationParametersStruct.
*/                                                                          /*
*/  } GC_CoverageParametersStruct;                                          /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Structure for solution found by coverage solver **************************
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_CoverageValueType     CoverageValueForCoverableArea;             /*
*/      GC_CoverageValueType     CoverageValueForEntireArea;                /*
*/      GC_InputVectorIndexType  Cardinality;                               /*
*/      GC_InputVectorIndexType* Indices;                                   /*
*/                                                                          /*
*/  } GC_SingleCoverageSolutionStruct;                                      /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Structure for set of solutions found by coverage solver ******************
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_AlgorithmType                 AlgorithmType;                     /* The used algorithm type (exhaustive or greedy).
*/                                                                          /*
*/      GC_LargestUnsignedIntegerType    NrOfUVMapValues;                   /* Total number of values stored in the UV maps (not only for valid target locations, but also for DONT CARE locations).
*/      cl_double                        MaximalPossibleCoverage;           /* Percentage of maximal possible coverage.
*/                                                                          /*
*/      GC_NrOfFoundSolutionsType        NrOfFoundSolutions;                /* Number of solutions found and stored.
*/      GC_SingleCoverageSolutionStruct* Solutions;                         /* Stored solutions.
*/      GC_InputVectorIndexType          IndexOfBestSolution;               /* Index of solution providing highest coverage.
*/                                                                          /*
*/      GC_LargestUnsignedIntegerType    MaximalTargetVectorEntry;          /* Maximal value in target vector.
*/                                                                          /*
*/      GC_LargestUnsignedIntegerType    TotalKernelExecutionTime;          /* Time for executing the main kernel. Required when computing the optimized settings.
*/                                                                          /*
*/      GC_GeneralParametersStruct*      GeneralParameters;                 /*
*/                                                                          /*
*/  } GC_CoverageSolutionStruct;                                            /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Structure for visualization settings *************************************
******************************************************************************
*/                                                                          /*
*/  typedef struct {                                                        /*
*/                                                                          /*
*/      GC_VisualizationTypeEnum VisualizationType;                         /*
*/      cl_double                UpperBoundFactor;                          /*
*/      cl_double                UpperEnvelopeThresholdFactor;              /*
*/      cl_double                LowerEnvelopeThresholdFactor;              /*
*/      cl_double                LowerBoundFactor;                          /*
*/                                                                          /*
*/  } GC_VisualizationSettingsStruct;                                       /*
*/                                                                          /*
*****************************************************************************/



/*****************************************************************************
*** Functions ****************************************************************
******************************************************************************
*/                                                                          /*
*   Main entry function                                                      *
*/                                                                          /*
*/  GC_BooleanType GC_ComputeSolution                                       /*
*/      (      GC_RendererParametersStruct* RendererParameters,             /*
*/             GC_CoverageParametersStruct* CoverageParameters,             /*
*/             GC_CoverageSolutionStruct*   CoverageSolution,               /*
*/             void                         (*Printf)(GC_CharacterType*,    /*
*/                                                    GC_MessageKind),      /*
*/       const GC_BooleanType*              ComputationStopRequested);      /*
*/                                                                          /*
******************************************************************************
*/                                                                          /*
*/  void GC_ReleaseParameterStructs                                         /*
*/      (      GC_RendererParametersStruct** RendererParameters,            /*
*/             GC_CoverageParametersStruct** CoverageParameters);           /*
*/                                                                          /*
******************************************************************************
*/                                                                          /*
*/  void GC_ResetCoverageSolutionStruct                                     /*
*/      (GC_CoverageSolutionStruct* CoverageSolution);                      /*
*/                                                                          /*
******************************************************************************
*/                                                                          /*
*/  void GC_GetFilePath                                                     /*
*/      (      GC_CharacterType*           String,                          /*
*/       const GC_GeneralParametersStruct* GeneralParameters,               /*
*/       const GC_NrOfObjectsType          ObjectID,                        /*
*/       const GC_FileTypeEnum             TypeOfFile);                     /*
*/                                                                          /*
******************************************************************************
*/                                                                          /*
*/  GC_BooleanType GC_CreateRGBAMaps                                        /*
*/      (const GC_CoverageSolutionStruct*     CoverageSolution,             /* Solution previously computed by means of function GC_ComputeSolution.
*/       const GC_InputVectorIndexType*       IndicesOfSensorsInSolution,   /* Indices of sensors in the union for which the png files shall be created.
*/       const GC_VisualizationSettingsStruct VisualizationSettings,        /* Settings required to synthesize the png files.
*/       const GC_CharacterType*              DirectoryPath,                /* Path to directory in which the png files shall be created.
*/       const GC_FileTypeEnum                TypeOfFile,                   /* Type of file (see also GC_FileTypeEnum).
*/             void                           (*Printf)(GC_CharacterType*,  /* Address of logging function.
*/                                                      GC_MessageKind),    /*
*/             GC_CharacterType**             LogInformation);              /* Address of string containing the logged information.
*/                                                                          /*
******************************************************************************
*/                                                                          /*
*/  GC_BooleanType GC_PerformOptimizationRun                                /*
*/      (const char*                                    MainDirectoryPath,  /* Path to the directory in which the file containing the optimal settings and the work directory are stored.
*/       const char*                                    WorkDirectoryName,  /* Name of the word directory.
*/             GC_CoverageOptimizationParametersStruct* BitPrecision,       /* Structure containing - after execution of the function - the optimal settings when solving a BGC problem.
*/             GC_CoverageOptimizationParametersStruct* BytePrecisionBinarySensors,   /* Structure containing - after execution of the function - the optimal settings when solving a DGC problem with binary sensors.
*/             GC_CoverageOptimizationParametersStruct* BytePrecisionNonBinarySensors,/* Structure containing - after execution of the function - the optimal settings when solving a DGC problem.
*/             void                                     (*Printf)(GC_CharacterType*,  /* Address of logging function.
*/                                                                GC_MessageKind),    /*
*/       const GC_BooleanType*                          ComputationStopRequested);    /* Indicates whether the computation shall be stopped.
*/                                                                          /*
*****************************************************************************/
#endif /* GC_H */