#ifndef GC_COVERAGE_SOLVER_H
#define GC_COVERAGE_SOLVER_H

/****************************************************************************/

#include "CyclicList.h"
#include "CoverageSolverInputData.h"

/****************************************************************************/
/* Type definitions *********************************************************/
/****************************************************************************/

typedef GC_LargestUnsignedIntegerType GC_NrOfOpenCLDevicesType;
typedef cl_uchar                      GC_KernelIndexType;
typedef cl_ulong                      GC_MaxCoverageValueType;
typedef cl_double                     GC_SensorDistanceType;

/****************************************************************************/
/* Enumerations *************************************************************/
/****************************************************************************/

typedef enum {

    GC_CHAR,
    GC_SHORT,
    GC_INT,
    GC_LONG,
    GC_NR_OF_INTEGER_TYPES

} GC_IntegerTypesEnum;

/****************************************************************************/
/* Structs ******************************************************************/
/****************************************************************************/

typedef struct {

    /* Info flags */
    GC_BooleanType                  OpenCLInfoModeSwitchedOn;

    /* Debugging flags */
    GC_BooleanType                  OpenCLDebugModeSwitchedOn;
    GC_BooleanType                  WriteKernelProgramSourceCodeToFile;
    GC_BooleanType                  LogKernelProgramBuildError;
    GC_BooleanType                  WriteKernelProgramBuildErrorLogToFile;

    /* Variable for storing OpenCL error */
    GC_OpenCLErrorType              Error;

    /* OpenCL context related variables */
    cl_context                      Context;
    cl_device_id                    DeviceID;
    cl_command_queue                CommandQueue;
    cl_device_id*                   IDsOfAvailableDevices;
    GC_CharacterType*               ExtensionString;
    GC_NrOfOpenCLDevicesType        NrOfAvailableDevices;
    GC_CharacterType                OpenCLVersionString[GC_MAX_INFO_STRING_LENGTH];
    cl_uint                         PreferredVectorCardinality[GC_NR_OF_INTEGER_TYPES];

    /* Program related variables */
    cl_program                      Program;
    GC_CharacterType*               ProgramBuildMessage;

    /* Kernel related variables */
    cl_kernel                       KernelIDs[GC_MAXIMAL_NR_OF_KERNELS];
    cl_ushort                       NrOfKernels;
    GC_CharacterType*               KernelIdentifiers[GC_MAXIMAL_NR_OF_KERNELS];
    cl_uint                         LocalWorkSize;                                   /* Current local work size. */
    cl_uint                         MaximalLocalWorkSizes[GC_MAXIMAL_NR_OF_KERNELS]; /* Local work sizes. */
    GC_LargestUnsignedIntegerType   CurrentGlobalWorkSize;                           /* Current global work size (needed for kernel execution). */
    GC_NrOfIterationsType           NrOfIterations;
    size_t                          DeviceMaximalWorkGroupSize;

    /* Profiling variables */
    GC_BooleanType                  ProfilingModeSwitchedOn;
    cl_event                        TimeProfilingEvent;
    GC_LargestUnsignedIntegerType   TotalKernelExecutionTime;

    /* Should actually use cl_ulong here according to OpenCL specification, */
    /* hence making sure that enough memory in case of 32 bit system.       */
#ifdef WIN32
    GC_LargestUnsignedIntegerType   TimeProfilingStartTime[2];
    GC_LargestUnsignedIntegerType   TimeProfilingStopTime[2];
#else /* 64 bit system */
    GC_LargestUnsignedIntegerType   TimeProfilingStartTime;
    GC_LargestUnsignedIntegerType   TimeProfilingStopTime;
#endif

    /* Graphics card dependent variables */
    GC_BooleanType                  UsePopCount;
    GC_BooleanType                  AMDPopCount;
    GC_CharacterType*               PopCountString;

    /* Buffer ID variables */
    cl_mem                          InputVectorsBufferID;
    cl_mem                          TargetVectorBufferID;
    cl_mem                          TotalNrOfElementsOutOfBoundBufferID;
    cl_mem                          IndicesBufferID;
    cl_mem                          WeightsBufferID;
    cl_mem                          ResultBufferID;

    /* Variables used for ping pong computation of best subset. */
    GC_NrOfIterationsType CurrentNrOfReducedElements;                         /* Current nr of elements left in buffer (needed for kernel execution). */
    GC_NrOfIterationsType CurrentOffsetBetweenResultElementAndIndexElement;   /* Current offset between result element and first index element. */

    /* Should actually use cl_ulong here according to OpenCL specification, */
    /* hence making sure that enough memory in case of 32 bit system.       */
#ifdef WIN32
    GC_LargestUnsignedIntegerType   DeviceGlobalMemorySize[2];
#else /* 64 bit system */
    GC_LargestUnsignedIntegerType   DeviceGlobalMemorySize;
#endif

    /* Should actually use cl_ulong here according to OpenCL specification, */
    /* hence making sure that enough memory in case of 32 bit system.       */
#ifdef WIN32
    GC_LargestUnsignedIntegerType   DeviceLocalMemorySize[2];
#else /* 64 bit system */
    GC_LargestUnsignedIntegerType   DeviceLocalMemorySize;
#endif

} GC_OpenCLVariablesStruct;

/****************************************************************************/

typedef struct {

    GC_InputVectorIndexType*        Indices;
    GC_InputVectorIndexType*        StoredIndices;

    GC_ResultElementType*           CompressedResults;
    GC_InputVectorIndexType         NrOfResultElements;                       /* Nr of elements of tuple stored in CompressedResults. */
    GC_BitIteratorType              Offset;                                   /* Number of bits required for storing one index of a input vector. */
    GC_InputVectorIndexType         OffsetBitMask;                            /* Bit mask consisting of Offset set bits. */

} GC_CoverageSolverIterationResultDataStruct;

/****************************************************************************/

typedef struct {

    GC_BooleanType                EnoughSolutionsFound;

    GC_NrOfFoundSolutionsType     NrOfFoundSolutions;
    GC_NrOfFoundSolutionsType     IndexOfListOfIndicesOfBestSolution;

    GC_NrOfFoundSolutionsType     MaxNrOfStoredSolutions;

    GC_ResultElementType*         CompressedResults;
    GC_ResultElementType*         TotalNrOfElementsOutOfBound;

    GC_ResultElementType          MinimalNrOfElementsOutOfBoundFoundSoFar;

    GC_CyclicListStruct*          ListOfIndicesPerSolution;
    
    /* Used only for greedy bit precision */
    GC_LargestUnsignedIntegerType NrOfRemainingNotCoveredTargetLocations;

} GC_CoverageSolverFoundSolutionsStruct;

/****************************************************************************/

typedef struct {

    GC_CharacterType              DebugDirectoryPath[GC_MAX_FILE_NAME_LENGTH];

} GC_CoverageSolverAuxiliaryVariablesStruct;

/****************************************************************************/

typedef struct {

    const GC_BooleanType*                      ComputationStopRequested;

    const GC_CoverageParametersStruct*         Parameters;

    const GC_CoverageInputDataStruct*          InputData;

    GC_StringSaverStruct                       ProgramSourceSaver;

    GC_StringSaverStruct*                      OutputLogger;

    GC_CoverageSolverAuxiliaryVariablesStruct  Aux;

    /* OpenCL related variables */
    GC_OpenCLVariablesStruct                   CL;

    /* Result related variables */
    GC_CoverageSolverIterationResultDataStruct IterationResults;
    GC_CoverageSolverFoundSolutionsStruct      FoundSolutions;

} GC_CoverageSolverStruct;

/****************************************************************************/
/* Functions ****************************************************************/
/****************************************************************************/

/* Code entry function ******************************************************/

GC_BooleanType SolveCoverageProblem
    (const GC_CoverageParametersStruct*     Parameters,
     const GC_CoverageInputDataStruct*      InputData,
           GC_CoverageSolutionStruct*       Solution,
     const GC_StringSaverStruct*            OutputLogger,
     const GC_BooleanType*                  ComputationStopRequested);

/* Initialization function **************************************************/

void InitializeSolver
    (      GC_CoverageSolverStruct*         Self,
     const GC_CoverageParametersStruct*     Parameters,
     const GC_CoverageInputDataStruct*      InputData,
     const GC_StringSaverStruct*            OutputLogger,
     const GC_BooleanType*                  ComputationStopRequested);

/* OpenCL setup functions ***************************************************/

void SetUpOpenCLEnvironment
    (      GC_CoverageSolverStruct*         Self);

void CreateOpenCLContext
    (      GC_CoverageSolverStruct*         Self);

void RetrieveOpenCLDevice
    (      GC_CoverageSolverStruct*         Self);

void RetrieveOpenCLCommandQueue
    (      GC_CoverageSolverStruct*         Self);

void RetrieveDeviceInformation
    (      GC_CoverageSolverStruct*         Self);

void RetrieveKernelSpecificInformation
    (      GC_CoverageSolverStruct*         Self);

void PerformCompatibilityCheckBetweenCodeAndOpenCLVersion
    (      GC_CoverageSolverStruct*         Self);

void PrintOpenCLInfo
    (      GC_CoverageSolverStruct*         Self);

/* OpenCL program source creation function **********************************/

void CreateOpenCLProgramHeader
    (      GC_CoverageSolverStruct*         Self);

void CreateComputeSumsOpenCLKernel
    (      GC_CoverageSolverStruct*         Self);

void CreateResetResultBufferOpenCLKernel
    (      GC_CoverageSolverStruct*         Self);

void CreateSubtractInputVectorOpenCLKernel
    (      GC_CoverageSolverStruct*         Self);

void CreateOpenCLProgramSource
    (      GC_CoverageSolverStruct*         Self);

void AddLUTCreationToProgramSource
    (      GC_CoverageSolverStruct*         Self,
     const GC_BooleanType                   MultiplyWithWeightValue);

void AppendAtomAddToProgramSource
    (      GC_CoverageSolverStruct*         Self,
     const GC_BooleanType                   UseLocalArray,
     const GC_BooleanType                   MultiplyWithValues);

/* OpenCL program creation function *****************************************/

void CreateOpenCLProgram
    (      GC_CoverageSolverStruct*         Self);

/* OpenCL buffer preparation functions **************************************/

void CreateBuffersOnDevice
    (      GC_CoverageSolverStruct*         Self);

void CreateInputVectorBuffers
    (      GC_CoverageSolverStruct*         Self);

void CreateIndicesBuffer
    (      GC_CoverageSolverStruct*         Self);

void CreateWeightVectorBuffer
    (      GC_CoverageSolverStruct*         Self);

void CreateResultBuffer
    (      GC_CoverageSolverStruct*         Self);

void CreateTotalNrOfElementsOutOfBoundBuffer
    (      GC_CoverageSolverStruct*         Self);

/* Function computing the maximal possible number of subset computations    */
/* before running out of memory.                                            */
void ComputeMaxNrOfIterations
    (      GC_CoverageSolverStruct*         Self);

/* Solving functions ********************************************************/

void Solve
    (      GC_CoverageSolverStruct*         Self);

void SolveExhaustively
    (      GC_CoverageSolverStruct*         Self);

void SolveGreedily
    (      GC_CoverageSolverStruct*         Self);

/* Kernel intialization and execution functions *****************************/

void PrepareAndExecuteKernel
     (      GC_CoverageSolverStruct*        Self,
      const GC_KernelIndexType              KernelID,
      const GC_InputVectorIndexType*        CurrentIndices,
      const GC_InputVectorIndexType         CurrentCardinality,
      const GC_NrOfIterationsType           NrOfCombinationsToCheck);

void UpdateIndices
    (const GC_CoverageParametersStruct*     Parameters,
           GC_InputVectorIndexType*         Indices,
           GC_InputVectorIndexType*         Cardinality);

void ResetResultBuffer
    (      GC_CoverageSolverStruct*         Self);

void ComputeCovers
     (      GC_CoverageSolverStruct*        Self,
      const GC_InputVectorIndexType         CurrentCardinality,
      const GC_InputVectorIndexType*        Indices,
      const GC_NrOfIterationsType           NrOfCombinationsToCheck);

void ComputeMaximalCoverage
    (      GC_CoverageSolverStruct*         Self,
     const GC_NrOfIterationsType            NrOfElements);

void IterativelyAddBestIndices
    (      GC_CoverageSolverStruct*         Self);

void StoreIndicesInOneList
    (      GC_CoverageSolverStruct*         Self);

void SubtractInputVector
    (      GC_CoverageSolverStruct*         Self);

/* Result retrieval functions ***********************************************/

void RetrieveResultsFromDevice
    (      GC_CoverageSolverStruct*         Self,
     const GC_NrOfIterationsType            NrOfCombinationsToCheck);

void ReadBackResult2
    (      GC_CoverageSolverStruct*         Self,
           cl_mem                           MemoryID);

/* Final result preparation function ****************************************/

void PerformConsistencyCheck
    (const GC_CoverageSolverStruct*         Self);

GC_InputVectorIndexType GetIndexOfBestSolution
    (const GC_CoverageSolverStruct*         Self);

GC_InputVectorIndexType ComputeBestIndex
    (      GC_CoverageSolverStruct*         Self);

void ExtractAndStoreValidSolutions
    (      GC_CoverageSolverStruct*         Self,
     const GC_NrOfIterationsType            IdOfFirstSolutionToCheck,
     const GC_NrOfIterationsType            IdOfLastSolutionToCheck,
     const GC_NrOfIterationsType            NrOfPossibleSolutions);

void ExtractIndicesAndAddToList
    (const GC_CoverageSolverStruct*         Self,
           GC_ResultElementType*            CompressedIndices,
           GC_CyclicListStruct*             List);

GC_BooleanType IsValidSolution
    (const GC_CoverageSolverStruct*         Self,
     const GC_ResultElementType             Element,
     const GC_ResultElementType             NrOfElementsOutOfBound);

void StoreAndPrintSolution
    (const GC_CoverageSolverStruct*         Self,
           GC_CoverageSolutionStruct*       Solution);
/*
void SortResultIndices
    (      GC_CoverageSolverStruct*         Self);
*/
/* Error checking / log functions *******************************************/

void CheckOpenCLErrors
    (      GC_CoverageSolverStruct*         Self,
     const GC_CharacterType*                Label);

void PrintProgramBuildMessage
    (      GC_CoverageSolverStruct*         Self);

/* Clean up function ********************************************************/

void CleanUpCoverageSolver
    (      GC_CoverageSolverStruct*         Self);

/* Helper functions *********************************************************/

void PrintCurrentSubsetInformation
    (      GC_CoverageSolverStruct*         Self,
     const GC_InputVectorIndexType          CurrentCardinality,
     const GC_InputVectorIndexType*         Indices);

GC_CharacterType* CLErrorMessage
    (const GC_OpenCLErrorType               Error);

GC_CharacterType* GetProgramSourceFromFile
    (const GC_CharacterType*                ProgramSourceFilePath);

/****************************************************************************/

#endif /* GC_COVERAGE_SOLVER_H */