#ifndef GC_RENDERER_H
#define GC_RENDERER_H

/****************************************************************************/

#include <GL/glew/glew.h>

#include "RendererParameters.h"

/****************************************************************************/
/* Global variables *********************************************************/
/****************************************************************************/

static jmp_buf RendererErrorHandlingEnvironment;

/*****************************************************************************
* Definitions ****************************************************************
*****************************************************************************/

#define GC_IS_OCCLUDER(x)         (x -> IDs[0] == GC_OCCLUDER_TRIANGLE_ID)

/*** XYZ ********************************************************************/

#define GC_XYZ_VERTEX_SHADER_NAME         "XYZVertexShader"

#define GC_XYZ_FRAGMENT_SHADER_NAME       "XYZFragmentShader"

#define GC_XYZ_MAP_TEXTURE_UNIFORM_STRING "XYZMapTexture"

#define GC_XYZ_IMAGE_FORMAT               GL_RED_INTEGER
#define GC_XYZ_IMAGE_BIT_DEPTH            8
#define GC_XYZ_IMAGE_INTERNAL_FORMAT      GL_R32UI
#define GC_XYZ_IMAGE_OUTPUT_TYPE          GL_UNSIGNED_INT
#define GC_XYZ_IMAGE_COLOR_TYPE           PNG_COLOR_TYPE_RGBA
#define GC_XYZ_IMAGE_NR_OF_CHANNELS       4

/*** GC_UV ******************************************************************/

#define GC_UV_VERTEX_SHADER_NAME          "UVVertexShader"
#define GC_UV_FRAGMENT_SHADER_NAME        "UVFragmentShader"

#define GC_UV_OLD_MAP_TEXTURE_UNIFORM_STRING "UVMapTexture"

#define GC_MAX_NR_OF_VIEWS               6

#define GC_TOTAL_NR_OF_TEXTURES          (GC_XYZ_NR_OF_TEXTURES + GC_UV_NR_OF_TEXTURES)
#define GC_MAX_NR_OF_TEXTURES_PER_SPACE  GC_UV_NR_OF_TEXTURES

#define GC_UV_IMAGE_FORMAT               GL_RED_INTEGER
#define GC_UV_IMAGE_BIT_DEPTH            8

/* 8 bit mode */
#define GC_UV_IMAGE_INTERNAL_FORMAT_8    GL_R8UI
#define GC_UV_IMAGE_OUTPUT_TYPE_8        GL_UNSIGNED_BYTE
#define GC_UV_IMAGE_COLOR_TYPE_8         PNG_COLOR_TYPE_GRAY
#define GC_UV_IMAGE_NR_OF_CHANNELS_8     1

/* 16 bit mode */
#define GC_UV_IMAGE_INTERNAL_FORMAT_16   GL_R16UI
#define GC_UV_IMAGE_OUTPUT_TYPE_16       GL_UNSIGNED_SHORT
#define GC_UV_IMAGE_COLOR_TYPE_16        PNG_COLOR_TYPE_GRAY_ALPHA
#define GC_UV_IMAGE_NR_OF_CHANNELS_16    2

/* 32 bit mode */
#define GC_UV_IMAGE_INTERNAL_FORMAT_32   GL_R32UI
#define GC_UV_IMAGE_OUTPUT_TYPE_32       GL_UNSIGNED_INT
#define GC_UV_IMAGE_COLOR_TYPE_32        PNG_COLOR_TYPE_RGBA
#define GC_UV_IMAGE_NR_OF_CHANNELS_32    4

/****************************************************************************/
/* Type definitions *********************************************************/
/****************************************************************************/

typedef cl_uint GC_OpenGLErrorType;

/****************************************************************************/
/* Enumerations *************************************************************/
/****************************************************************************/

typedef enum {

    GC_VERTEX_SHADER,
    GC_FRAGMENT_SHADER,
    GC_NR_OF_SHADERS_PER_PROGRAM

} GC_ShaderTypeEnum;

/****************************************************************************/

typedef enum {

    GC_OBJECTS_VERTICES_XYZ,
    GC_OBJECTS_IDS,
    GC_OBJECTS_VERTICES_UV,
    GC_OBJECTS_NORMALS,
    GC_NR_OF_BUFFERS_PER_OBJECT

} GC_VertexArrayEnum;

/****************************************************************************/

typedef enum {

     GC_XYZ_MAP_TEXTURE,
     GC_XYZ_NR_OF_TEXTURES,
     GC_UV_NEW_MAP_TEXTURE = GC_XYZ_NR_OF_TEXTURES,
     GC_UV_OLD_MAP_TEXTURE,
     GC_UV_NR_OF_TEXTURES  = GC_UV_OLD_MAP_TEXTURE

} GC_MapTextureEnum;

/****************************************************************************/
/* Structs ******************************************************************/
/****************************************************************************/

/* Shader program related variables *****************************************/

typedef struct {

          GC_CharacterType*             Name;
    const GC_LargestUnsignedIntegerType Type;
          GC_CharacterType*             Source;

} GC_ShaderStruct;

/* Shader program related variables *****************************************/

typedef struct {

    GLuint             ProgramID;
    GLuint             ShaderIDs[GC_NR_OF_SHADERS_PER_PROGRAM];

    GLenum             VertexXYZPositionAttribute;
    GLenum             VertexIdAttribute;
    GLenum             ViewUniform;
    GLenum             ProjectionUniform;

    GC_MatrixEntryType ProjectionMatrix[GC_MATRIX_DIMENSION][GC_MATRIX_DIMENSION];

    /* Variables used only by UV shader */
    GLenum             VertexUVPositionAttribute;
    GLenum             VertexNormalAttribute;
    GLenum             FactorUniform;
    GLenum             PrecisionUniform;
    GLenum             BinarySensorsUniform;
    GLenum             IsFirstViewUniform;
    GLenum             SideLengthOfXYZMapUniform;
    GLenum             NeighborhoodRadiusUniform;

} GC_ShaderDataStruct;

/* Render-framebuffer and renderbuffer related variables ********************/

typedef struct {

    GLenum              FrameBufferID;
    GLenum              DepthRenderBufferID;
    GLenum              InternalFormat;
    GLenum              Format;
    GLenum              OutputType;
    GC_ColorTypeType    ColorType;
    GC_BitDepthType     BitDepth;
    GC_NrOfChannelsType NrOfChannels;

} GC_RenderBufferStruct;

/* Renderer auxiliary variables struct **************************************/


typedef struct {

    GC_BooleanType                OpenGLDebugModeSwitchedOn;
    GC_LargestUnsignedIntegerType NrOfTargetsAndOccluders;
    GC_BitIteratorType            NrOfBytesRequiredPerUVMapValue;
    GC_CharacterType              DebugDirectoryPath[GC_MAX_FILE_NAME_LENGTH];

} GC_RendererAuxiliaryVariablesStruct;

/* Renderer struct **********************************************************/

typedef struct {

    const GC_BooleanType*               ComputationStopRequested;

    GC_BooleanType                      HasBeenCleanedUp;

    GC_RendererParametersStruct*        Parameters;

    GLenum                              Error;
    GC_PngImageStruct                   PngImage[GC_NR_OF_SPACES];
    GC_StringSaverStruct*               OutputLogger;

    GC_RendererAuxiliaryVariablesStruct Aux;

    /************************************************************************/
    /* Shader program related variables *************************************/
    /************************************************************************/

    GC_ShaderDataStruct                 ShaderData[GC_NR_OF_SPACES];

    /************************************************************************/
    /* Vertex array related variables ***************************************/
    /************************************************************************/

    GLuint*                             VertexArrayID;
    GLuint*                             VertexArrayBufferID;

    /************************************************************************/
    /* Render-framebuffer and renderbuffer related variables ****************/
    /************************************************************************/

    GC_RenderBufferStruct               RenderBufferData[GC_NR_OF_SPACES];

    /************************************************************************/
    /* Texture related variables ********************************************/
    /************************************************************************/

    GLuint                              TextureIDs[GC_TOTAL_NR_OF_TEXTURES];
    GLuint                              TextureUniforms[GC_TOTAL_NR_OF_TEXTURES];
    GLenum                              ColorAttachments[GC_TOTAL_NR_OF_TEXTURES];

    /************************************************************************/
    /* Data buffer related variables ****************************************/
    /************************************************************************/

    GC_XYZMapValueType*                 XYZMapData;

    GC_UVMapsStruct                     UVMaps;

} GC_RendererStruct;

/****************************************************************************/
/* Functions ****************************************************************/
/****************************************************************************/

GC_CharacterType* ReadShaderSource
    (      GC_StringSaverStruct*            OutputLogger,
     const GC_CharacterType*                ShaderFileName);

/* Code entry function ******************************************************/

GC_BooleanType RenderUVMaps
    (const GC_RendererParametersStruct*     Parameters,
     const GC_StringSaverStruct*            OutputLogger,
     const GC_LargestUnsignedIntegerType*   NrOfUVMapValues,
     const GC_BooleanType*                  ComputationStopRequested);

/****************************************************************************/

void AssociateShaderAttributes
    (      GC_RendererStruct*               Self,
     const GC_Space                         CurrentSpace);

void BindVertexArrayAndBuffers
    (      GC_RendererStruct*               Self,
     const GC_NrOfObjectsType               ObjectID,
     const GC_Space                         CurrentSpace);

void ComputeMap
    (      GC_RendererStruct*               Self,
     const GC_Space                         CurrentSpace,
     const GC_InputVectorIndexType          SensorID,
     const GC_NrOfViewsType                 ViewID,
     const GC_NrOfObjectsType               ObjectID);

void CleanUpRenderer
    (      GC_RendererStruct*               Self);

void ComputeXYZandUVMaps
    (      GC_RendererStruct*               Self,
           GC_LargestUnsignedIntegerType*   NrOfUVMapValues);

void CopyTextureDataFromDeviceToHost
    (      GC_RendererStruct*               Self,
     const GC_ImageSizeType                 BlockLength,
     const GC_ImageSizeType                 Width,
     const GC_ImageSizeType                 Height,
     const GC_Space                         CurrentSpace);

void CopyTextureDataFromDeviceToHostAndPotentiallyWriteToFile
    (      GC_RendererStruct*               Self,
     const GC_NrOfObjectsType               ObjectID,
     const GC_BooleanType                   StoreImage,
     const GC_Space                         CurrentSpace);

void CopyTextureDataFromHostToDevice
    (      GC_RendererStruct*               Self,
     const GC_NrOfObjectsType               ObjectID);

void PrepareMapComputation
    (      GC_RendererStruct*               Self,
     const GC_Space                         CurrentSpace,
     const GC_InputVectorIndexType          SensorID,
     const GC_NrOfViewsType                 ViewID);

void CreateAndCompileShaders
    (      GC_RendererStruct*               Self,
     const GC_Space                         CurrentSpace);

void CreateAndLinkShaderProgramsOnDevice
    (      GC_RendererStruct*               Self);

void Draw
    (      GC_RendererStruct*               Self,
     const GC_NrOfObjectsType               ObjectID);

void SetUpOpenGLEnvironment
    (const GC_StringSaverStruct*            OutputLogger);

void InitializeMapDataBuffersOnHost
    (      GC_RendererStruct*               Self);

void InitializeRenderer
    (      GC_RendererStruct*               Self,
     const GC_RendererParametersStruct*     Parameters,
     const GC_StringSaverStruct*            OutputLogger,
     const GC_BooleanType*                  ComputationStopRequested);

void InitializeTexturesAndDepthRenderBufferOnDevice
    (      GC_RendererStruct*               Self);

void InitializeVertexArraysOnDevice
    (      GC_RendererStruct*               Self);

void LinkShaderProgram
    (      GC_RendererStruct*               Self,
     const GC_Space                         CurrentSpace);

void OpenImage
    (      GC_RendererStruct*               Self,
     const GC_ImageInitializationType       FileType,
     const GC_InputVectorIndexType          SensorID,
     const GC_NrOfViewsType                 ViewID,
     const GC_NrOfObjectsType               ObjectID,
     const GC_Space                         CurrentSpace);

void RenderMap
    (      GC_RendererStruct*               Self,
     const GC_NrOfObjectsType               ObjectID,
     const GC_Space                         CurrentSpace);

void WriteTextureDataToFile
    (      GC_RendererStruct*               Self,
     const GC_ImageSizeType                 BlockLength,
     const GC_ImageSizeType                 Width,
     const GC_Space                         CurrentSpace);

/* Error checking functions *************************************************/

void CheckOpenGLErrors
    (      GC_RendererStruct*               Self,
     const GC_CharacterType*                ErrorMessage);

GC_CharacterType* OpenGLErrorMessage
    (const GC_OpenGLErrorType               Error);

/****************************************************************************/

#endif /* GC_RENDERER_H */